from .core import hints

Box = hints.Box
Contour = hints.Contour
Empty = hints.Empty
Linear = hints.Linear
Maybe = hints.Maybe
Mix = hints.Mix
Multipoint = hints.Multipoint
Multipolygon = hints.Multipolygon
Multisegment = hints.Multisegment
Point = hints.Point
Polygon = hints.Polygon
Scalar = hints.Scalar
Segment = hints.Segment
Shaped = hints.Shaped
Box.__module__ = __name__
Contour.__module__ = __name__
Empty.__module__ = __name__
Mix.__module__ = __name__
Multipoint.__module__ = __name__
Multipolygon.__module__ = __name__
Multisegment.__module__ = __name__
Point.__module__ = __name__
Polygon.__module__ = __name__
Segment.__module__ = __name__



from reprit.base import generate_repr

from ground.core.hints import (QuaternaryPointFunction,
                               Scalar)
from .exact import (cross as exact_cross,
                    dot as exact_dot)
from .plain import (cross as plain_cross,
                    dot as plain_dot)
from .robust import (cross as robust_cross,
                     dot as robust_dot)

QuaternaryFunction = QuaternaryPointFunction[Scalar]


class Context:
    __slots__ = '_cross_product', '_dot_product'

    def __init__(self,
                 cross_product: QuaternaryFunction,
                 dot_product: QuaternaryFunction) -> None:
        self._cross_product, self._dot_product = cross_product, dot_product

    __repr__ = generate_repr(__init__,
                             with_module_name=True)

    @property
    def cross_product(self) -> QuaternaryFunction:
        return self._cross_product

    @property
    def dot_product(self) -> QuaternaryFunction:
        return self._dot_product


exact_context = Context(cross_product=exact_cross.multiply,
                        dot_product=exact_dot.multiply)
plain_context = Context(cross_product=plain_cross.multiply,
                        dot_product=plain_dot.multiply)
robust_context = Context(cross_product=robust_cross.multiply,
                         dot_product=robust_dot.multiply)



from typing import Type

from ground.core.arithmetic import robust_divide
from ground.core.hints import (Point,
                               QuaternaryPointFunction,
                               TernaryPointFunction)
from ground.core.vector.plain import cross


def intersect(first_start: Point,
              first_end: Point,
              second_start: Point,
              second_end: Point,
              point_cls: Type[Point],
              containment_checker: TernaryPointFunction[bool],
              cross_product: QuaternaryPointFunction = cross.multiply
              ) -> Point:
    if containment_checker(first_start, first_end, second_start):
        return second_start
    elif containment_checker(first_start, first_end, second_end):
        return second_end
    elif containment_checker(second_start, second_end, first_start):
        return first_start
    elif containment_checker(second_start, second_end, first_end):
        return first_end
    else:
        scale = robust_divide(cross_product(first_start, second_start,
                                            second_start, second_end),
                              cross_product(first_start, first_end,
                                            second_start, second_end))
        return point_cls(first_start.x + (first_end.x - first_start.x) * scale,
                         first_start.y + (first_end.y - first_start.y) * scale)



from typing import Type

from shewchuk import Expansion

from ground.core.enums import Relation
from ground.core.hints import (Box,
                               Point,
                               QuaternaryPointFunction,
                               Scalar,
                               Segment)
from ground.core.primitive import square
from .segment import (point_squared_distance as segment_point_squared_distance,
                      segment_squared_distance
                      as segment_segment_squared_distance)


def point_squared_distance(box: Box, point: Point) -> Scalar:
    dx, dy = (_linear_interval_distance(box.min_x, box.max_x, point.x),
              _linear_interval_distance(box.min_y, box.max_y, point.y))
    return square(dx) + square(dy)


def segment_squared_distance(box: Box,
                             segment: Segment,
                             dot_producer: QuaternaryPointFunction[Scalar],
                             segments_relater
                             : QuaternaryPointFunction[Relation],
                             point_cls: Type[Point]) -> Scalar:
    segment_start, segment_end = segment.start, segment.end
    min_x, min_y, max_x, max_y = box.min_x, box.min_y, box.max_x, box.max_y
    return (0
            if ((min_x <= segment_start.x <= max_x
                 and min_y <= segment_start.y <= max_y)
                or (min_x <= segment_end.x <= max_x
                    and min_y <= segment_end.y <= max_y))
            else
            ((segment_point_squared_distance(segment_start, segment_end,
                                             point_cls(min_x, min_y),
                                             dot_producer)
              if min_y == max_y
              else segment_segment_squared_distance(
                    segment_start, segment_end, point_cls(min_x, min_y),
                    point_cls(min_x, max_y), dot_producer, segments_relater))
             if min_x == max_x
             else (segment_segment_squared_distance(
                    segment_start, segment_end, point_cls(min_x, min_y),
                    point_cls(max_x, min_y), dot_producer, segments_relater)
                   if min_y == max_y
                   else _non_degenerate_segment_squared_distance(
                    max_x, max_y, min_x, min_y, segment_start, segment_end,
                    dot_producer, segments_relater, point_cls))))


def _linear_interval_distance(min_coordinate: Scalar,
                              max_coordinate: Scalar,
                              coordinate: Scalar) -> Expansion:
    return (Expansion(min_coordinate, -coordinate)
            if coordinate < min_coordinate
            else (Expansion(coordinate, -max_coordinate)
                  if coordinate > max_coordinate
                  else Expansion()))


def _non_degenerate_segment_squared_distance(
        max_x: Scalar,
        max_y: Scalar,
        min_x: Scalar,
        min_y: Scalar,
        segment_start: Point,
        segment_end: Point,
        dot_producer: QuaternaryPointFunction[Scalar],
        segments_relater: QuaternaryPointFunction[Relation],
        point_cls: Type[Point]) -> Scalar:
    bottom_left, bottom_right = (point_cls(min_x, min_y),
                                 point_cls(max_x, min_y))
    bottom_side_distance = segment_segment_squared_distance(
            segment_start, segment_end, bottom_left, bottom_right,
            dot_producer, segments_relater)
    if not bottom_side_distance:
        return bottom_side_distance
    top_right = point_cls(max_x, max_y)
    right_side_distance = segment_segment_squared_distance(
            segment_start, segment_end, bottom_right, top_right, dot_producer,
            segments_relater)
    if not right_side_distance:
        return right_side_distance
    top_left = point_cls(min_x, max_y)
    top_side_distance = segment_segment_squared_distance(
            segment_start, segment_end, top_left, top_right, dot_producer,
            segments_relater)
    if not top_side_distance:
        return top_side_distance
    left_side_distance = segment_segment_squared_distance(
            segment_start, segment_end, bottom_left, top_left, dot_producer,
            segments_relater)
    return (left_side_distance
            and min(bottom_side_distance, right_side_distance,
                    top_side_distance, left_side_distance))



from typing import (Sequence,
                    Tuple,
                    Type)

from cfractions import Fraction
from shewchuk import Expansion

from ground.core.hints import (Multipolygon,
                               Point,
                               Polygon)
from .polygon import centroid_components as polygon_centroid_components


def centroid(multipolygon: Multipolygon,
             point_cls: Type[Point],
             third: Fraction = Fraction(1, 3)) -> Point:
    x_numerator, y_numerator, double_area = centroid_components(
            multipolygon.polygons)
    inverted_denominator = third / double_area
    return point_cls(x_numerator * inverted_denominator,
                     y_numerator * inverted_denominator)


def centroid_components(polygons: Sequence[Polygon]
                        ) -> Tuple[Expansion, Expansion, Expansion]:
    iterator = iter(polygons)
    polygon = next(iterator)
    x_numerator, y_numerator, double_area = polygon_centroid_components(
            polygon.border, polygon.holes)
    for polygon in iterator:
        (polygon_x_numerator, polygon_y_numerator,
         polygon_double_area) = polygon_centroid_components(polygon.border,
                                                            polygon.holes)
        x_numerator, y_numerator, double_area = (
            x_numerator + polygon_x_numerator,
            y_numerator + polygon_y_numerator,
            double_area + polygon_double_area)
    return x_numerator, y_numerator, double_area



from typing import (Callable,
                    Type)

from ground.core.hints import (Multisegment,
                               Point,
                               Scalar)


def centroid(multisegment: Multisegment,
             point_cls: Type[Point],
             sqrt: Callable[[Scalar], Scalar]) -> Point:
    accumulated_x = accumulated_y = accumulated_length = 0
    for segment in multisegment.segments:
        start, end = segment.start, segment.end
        length = sqrt((end.x - start.x) ** 2 + (end.y - start.y) ** 2)
        accumulated_x += (start.x + end.x) * length
        accumulated_y += (start.y + end.y) * length
        accumulated_length += length
    inverted_divisor = 1 / (2 * accumulated_length)
    return point_cls(accumulated_x * inverted_divisor,
                     accumulated_y * inverted_divisor)



from typing import (Tuple,
                    Type)

from shewchuk import Expansion

from ground.core.hints import (Point,
                               Scalar)


def rotate_point_around_origin(point: Point,
                               cosine: Scalar,
                               sine: Scalar,
                               point_cls: Type[Point]) -> Point:
    cosine, sine = Expansion(cosine), Expansion(sine)
    return point_cls(cosine * point.x - sine * point.y,
                     sine * point.x + cosine * point.y)


def rotate_translate_point(point: Point,
                           cosine: Scalar,
                           sine: Scalar,
                           step_x: Scalar,
                           step_y: Scalar,
                           point_cls: Type[Point]) -> Point:
    cosine, sine = Expansion(cosine), Expansion(sine)
    return point_cls(cosine * point.x - sine * point.y + step_x,
                     sine * point.x + cosine * point.y + step_y)


def point_to_step(point: Point,
                  cosine: Scalar,
                  sine: Scalar) -> Tuple[Scalar, Scalar]:
    cosine, sine = Expansion(cosine), Expansion(sine)
    return (point.x - (cosine * point.x - sine * point.y),
            point.y - (sine * point.x + cosine * point.y))



from shewchuk import Expansion

from ground.core.enums import Relation
from ground.core.hints import (Point,
                               QuaternaryPointFunction,
                               Scalar)
from ground.core.primitive import square


def point_squared_distance(start: Point,
                           end: Point,
                           point: Point,
                           dot_producer: QuaternaryPointFunction[Scalar]
                           ) -> Scalar:
    segment_squared_norm = dot_producer(start, end, start, end)
    point_end_projection = dot_producer(point, end, start, end)
    start_point_projection = dot_producer(start, point, start, end)
    start_factor = ((point_end_projection
                     if point_end_projection > 0
                     else 0)
                    if point_end_projection < segment_squared_norm
                    else segment_squared_norm)
    end_factor = ((start_point_projection
                   if start_point_projection > 0
                   else 0)
                  if start_point_projection < segment_squared_norm
                  else segment_squared_norm)
    return ((square(segment_squared_norm * Expansion(start.x, -point.x)
                    + end_factor * Expansion(end.x, -start.x))
             + square(segment_squared_norm * Expansion(start.y, -point.y)
                      + end_factor * Expansion(end.y, -start.y))
             if (abs(segment_squared_norm - end_factor)
                 < abs(segment_squared_norm - start_factor))
             else (square(segment_squared_norm * Expansion(end.x, -point.x)
                          + start_factor * Expansion(start.x, -end.x))
                   + square(segment_squared_norm * Expansion(end.y, -point.y)
                            + start_factor * Expansion(start.y, -end.y))))
            / square(segment_squared_norm))


def segment_squared_distance(first_start: Point,
                             first_end: Point,
                             second_start: Point,
                             second_end: Point,
                             dot_producer: QuaternaryPointFunction[Scalar],
                             segments_relater
                             : QuaternaryPointFunction[Relation]) -> Scalar:
    return (min(point_squared_distance(first_start, first_end, second_start,
                                       dot_producer),
                point_squared_distance(first_start, first_end, second_end,
                                       dot_producer),
                point_squared_distance(second_start, second_end, first_start,
                                       dot_producer),
                point_squared_distance(second_start, second_end, first_end,
                                       dot_producer))
            if segments_relater(first_start, first_end, second_start,
                                second_end) is Relation.DISJOINT
            else 0)



from reprit.base import generate_repr

from ground.core.enums import Location
from ground.core.hints import QuaternaryPointFunction
from .exact import point_point_point as exact_point_point_point
from .plain import point_point_point as plain_point_point_point
from .robust import point_point_point as robust_point_point_point

PointPointPointLocator = QuaternaryPointFunction[Location]


class Context:
    __slots__ = '_point_point_point_test',

    def __init__(self, point_point_point_test: PointPointPointLocator
                 ) -> None:
        self._point_point_point_test = point_point_point_test

    __repr__ = generate_repr(__init__,
                             with_module_name=True)

    @property
    def point_point_point_locator(self) -> PointPointPointLocator:
        return self._point_point_point_test


exact_context = Context(exact_point_point_point.test)
plain_context = Context(plain_point_point_point.test)
robust_context = Context(robust_point_point_point.test)



from ground.core.enums import Location
from ground.core.hints import (Point,
                               Scalar)
from ground.core.primitive import (rationalize,
                                   to_sign)


def test(point: Point,
         first: Point,
         second: Point,
         third: Point) -> Scalar:
    point_x, point_y = rationalize(point.x), rationalize(point.y)
    first_dx, first_dy = (rationalize(first.x) - point_x,
                          rationalize(first.y) - point_y)
    second_dx, second_dy = (rationalize(second.x) - point_x,
                            rationalize(second.y) - point_y)
    third_dx, third_dy = (rationalize(third.x) - point_x,
                          rationalize(third.y) - point_y)
    return Location(1
                    + to_sign((first_dx * first_dx + first_dy * first_dy)
                              * (second_dx * third_dy - second_dy * third_dx)
                              - (second_dx * second_dx + second_dy * second_dy)
                              * (first_dx * third_dy - first_dy * third_dx)
                              + (third_dx * third_dx + third_dy * third_dy)
                              * (first_dx * second_dy - first_dy * second_dx)))



from typing import (Sequence,
                    Tuple,
                    Type)

from cfractions import Fraction
from shewchuk import Expansion

from ground.core.hints import (Contour,
                               Point,
                               Polygon)
from .region import centroid_components as region_centroid_components


def centroid(polygon: Polygon,
             point_cls: Type[Point],
             third: Fraction = Fraction(1, 3)) -> Point:
    x_numerator, y_numerator, double_area = centroid_components(polygon.border,
                                                                polygon.holes)
    inverted_denominator = third / double_area
    return point_cls(x_numerator * inverted_denominator,
                     y_numerator * inverted_denominator)


def centroid_components(border: Contour,
                        holes: Sequence[Contour]
                        ) -> Tuple[Expansion, Expansion, Expansion]:
    x_numerator, y_numerator, double_area = region_centroid_components(
            border.vertices)
    for hole in holes:
        (hole_x_numerator, hole_y_numerator,
         hole_double_area) = region_centroid_components(hole.vertices)
        x_numerator, y_numerator, double_area = (
            x_numerator + hole_x_numerator, y_numerator + hole_y_numerator,
            double_area + hole_double_area)
    return x_numerator, y_numerator, double_area



from enum import (IntEnum,
                  unique)


class Base(IntEnum):
    def __repr__(self) -> str:
        return type(self).__qualname__ + '.' + self.name


@unique
class Kind(Base):
    """
    Represents kinds of angles
    based on their degrees value in range ``[0, 180]``.
    """
    #: ``(90, 180]`` degrees
    OBTUSE = -1
    #: ``90`` degrees
    RIGHT = 0
    #: ``[0, 90)`` degrees
    ACUTE = 1


@unique
class Location(Base):
    """
    Represents kinds of locations in which point can be relative to geometry.
    """
    #: point lies in the exterior of the geometry
    EXTERIOR = 0
    #: point lies on the boundary of the geometry
    BOUNDARY = 1
    #: point lies in the interior of the geometry
    INTERIOR = 2


@unique
class Orientation(Base):
    """
    Represents kinds of angle orientations.
    """
    #: in the same direction as a clock's hands
    CLOCKWISE = -1
    #: to the top and then to the bottom or vice versa
    COLLINEAR = 0
    #: opposite to clockwise
    COUNTERCLOCKWISE = 1


@unique
class Relation(Base):
    """
    Represents kinds of relations in which geometries can be.
    Order of members assumes that conditions for previous ones do not hold.
    """
    #: intersection is empty
    DISJOINT = 0
    #: intersection is a strict subset of each of the geometries,
    #: has dimension less than at least of one of the geometries
    #: and only boundaries intersect, but do not cross
    TOUCH = 1
    #: intersection is a strict subset of each of the geometries,
    #: has dimension less than at least of one of the geometries,
    #: one of the geometries lies in interior & exterior of the other geometry
    #: or boundaries cross
    CROSS = 2
    #: intersection is a strict subset of each of the geometries
    #: and has the same dimension as geometries
    OVERLAP = 3
    #: interior of the geometry is a superset of the other
    COVER = 4
    #: boundary of the geometry contains
    #: at least one boundary point of the other, but not all,
    #: interior of the geometry contains other points of the other
    ENCLOSES = 5
    #: geometry is a strict superset of the other
    #: and interior/boundary of the geometry is a superset
    #: of interior/boundary of the other
    COMPOSITE = 6
    #: geometries are equal
    EQUAL = 7
    #: geometry is a strict subset of the other
    #: and interior/boundary of the geometry is a subset
    #: of interior/boundary of the other
    COMPONENT = 8
    #: at least one boundary point of the geometry
    #: lies on the boundary of the other, but not all,
    #: other points of the geometry lie in the interior of the other
    ENCLOSED = 9
    #: geometry is a subset of the interior of the other
    WITHIN = 10

    @property
    def complement(self) -> 'Relation':
        if self is Relation.COVER:
            return Relation.WITHIN
        elif self is Relation.ENCLOSES:
            return Relation.ENCLOSED
        elif self is Relation.COMPOSITE:
            return Relation.COMPONENT
        elif self is Relation.COMPONENT:
            return Relation.COMPOSITE
        elif self is Relation.ENCLOSED:
            return Relation.ENCLOSES
        elif self is Relation.WITHIN:
            return Relation.COVER
        else:
            return self



from ground.core.enums import Location
from ground.core.hints import Point
from ground.core.primitive import to_sign


def test(point: Point, first: Point, second: Point, third: Point) -> Location:
    first_dx, first_dy = first.x - point.x, first.y - point.y
    second_dx, second_dy = second.x - point.x, second.y - point.y
    third_dx, third_dy = third.x - point.x, third.y - point.y
    return Location(1
                    + to_sign((first_dx * first_dx + first_dy * first_dy)
                              * (second_dx * third_dy - second_dy * third_dx)
                              - (second_dx * second_dx + second_dy * second_dy)
                              * (first_dx * third_dy - first_dy * third_dx)
                              + (third_dx * third_dx + third_dy * third_dy)
                              * (first_dx * second_dy - first_dy * second_dx)))



from typing import (Callable,
                    Tuple,
                    Type)

from reprit.base import generate_repr

from ground.core.hints import (Contour,
                               Multipoint,
                               Multipolygon,
                               Multisegment,
                               Point,
                               Polygon,
                               Scalar,
                               Segment)
from . import (exact,
               plain,
               robust)

PointRotatorAroundOrigin = Callable[[Point, Scalar, Scalar, Type[Point]],
                                    Point]
PointTranslatingRotator = Callable[[Point, Scalar, Scalar, Scalar, Scalar,
                                    Type[Point]], Point]
PointStep = Callable[[Point, Scalar, Scalar], Tuple[Scalar, Scalar]]


class Context:
    @property
    def point_to_step(self) -> PointStep:
        return self._point_to_step

    @property
    def rotate_point_around_origin(self) -> PointRotatorAroundOrigin:
        return self._rotate_point_around_origin

    @property
    def rotate_translate_point(self) -> PointTranslatingRotator:
        return self._rotate_translate_point

    def rotate_contour_around_origin(self,
                                     contour: Contour,
                                     cosine: Scalar,
                                     sine: Scalar,
                                     contour_cls: Type[Contour],
                                     point_cls: Type[Point]) -> Contour:
        return contour_cls([self.rotate_point_around_origin(point, cosine,
                                                            sine, point_cls)
                            for point in contour.vertices])

    def rotate_multipoint_around_origin(self,
                                        multipoint: Multipoint,
                                        cosine: Scalar,
                                        sine: Scalar,
                                        multipoint_cls: Type[Multipoint],
                                        point_cls: Type[Point]) -> Multipoint:
        return multipoint_cls([self.rotate_point_around_origin(point, cosine,
                                                               sine, point_cls)
                               for point in multipoint.points])

    def rotate_multipolygon_around_origin(self,
                                          multipolygon: Multipolygon,
                                          cosine: Scalar,
                                          sine: Scalar,
                                          contour_cls: Type[Contour],
                                          multipolygon_cls: Type[Multipolygon],
                                          point_cls: Type[Point],
                                          polygon_cls: Type[Polygon]
                                          ) -> Multipolygon:
        return multipolygon_cls(
                [self.rotate_polygon_around_origin(polygon, cosine, sine,
                                                   contour_cls, point_cls,
                                                   polygon_cls)
                 for polygon in multipolygon.polygons])

    def rotate_multisegment_around_origin(self,
                                          multisegment: Multisegment,
                                          cosine: Scalar,
                                          sine: Scalar,
                                          multisegment_cls: Type[Multisegment],
                                          point_cls: Type[Point],
                                          segment_cls: Type[Segment]
                                          ) -> Multisegment:
        return multisegment_cls(
                [self.rotate_segment_around_origin(segment, cosine, sine,
                                                   point_cls, segment_cls)
                 for segment in multisegment.segments])

    def rotate_polygon_around_origin(self,
                                     polygon: Polygon,
                                     cosine: Scalar,
                                     sine: Scalar,
                                     contour_cls: Type[Contour],
                                     point_cls: Type[Point],
                                     polygon_cls: Type[Polygon]) -> Polygon:
        return polygon_cls(
                self.rotate_contour_around_origin(polygon.border, cosine, sine,
                                                  contour_cls, point_cls),
                [self.rotate_contour_around_origin(hole, cosine, sine,
                                                   contour_cls, point_cls)
                 for hole in polygon.holes])

    def rotate_segment_around_origin(self,
                                     segment: Segment,
                                     cosine: Scalar,
                                     sine: Scalar,
                                     point_cls: Type[Point],
                                     segment_cls: Type[Segment]) -> Segment:
        return segment_cls(self.rotate_point_around_origin(segment.start,
                                                           cosine, sine,
                                                           point_cls),
                           self.rotate_point_around_origin(segment.end,
                                                           cosine, sine,
                                                           point_cls))

    def rotate_translate_contour(self,
                                 contour: Contour,
                                 cosine: Scalar,
                                 sine: Scalar,
                                 step_x: Scalar,
                                 step_y: Scalar,
                                 contour_cls: Type[Contour],
                                 point_cls: Type[Point]) -> Contour:
        return contour_cls([self.rotate_translate_point(point, cosine, sine,
                                                        step_x, step_y,
                                                        point_cls)
                            for point in contour.vertices])

    def rotate_translate_multipoint(self,
                                    multipoint: Multipoint,
                                    cosine: Scalar,
                                    sine: Scalar,
                                    step_x: Scalar,
                                    step_y: Scalar,
                                    multipoint_cls: Type[Multipoint],
                                    point_cls: Type[Point]) -> Multipoint:
        return multipoint_cls([self.rotate_translate_point(point, cosine, sine,
                                                           step_x, step_y,
                                                           point_cls)
                               for point in multipoint.points])

    def rotate_translate_multipolygon(self,
                                      multipolygon: Multipolygon,
                                      cosine: Scalar,
                                      sine: Scalar,
                                      step_x: Scalar,
                                      step_y: Scalar,
                                      contour_cls: Type[Contour],
                                      multipolygon_cls: Type[Multipolygon],
                                      point_cls: Type[Point],
                                      polygon_cls: Type[Polygon]
                                      ) -> Multipolygon:
        return multipolygon_cls(
                [self.rotate_translate_polygon(polygon, cosine, sine, step_x,
                                               step_y, contour_cls, point_cls,
                                               polygon_cls)
                 for polygon in multipolygon.polygons])

    def rotate_translate_multisegment(self,
                                      multisegment: Multisegment,
                                      cosine: Scalar,
                                      sine: Scalar,
                                      step_x: Scalar,
                                      step_y: Scalar,
                                      multisegment_cls: Type[Multisegment],
                                      point_cls: Type[Point],
                                      segment_cls: Type[Segment]
                                      ) -> Multisegment:
        return multisegment_cls(
                [self.rotate_translate_segment(segment, cosine, sine, step_x,
                                               step_y, point_cls, segment_cls)
                 for segment in multisegment.segments])

    def rotate_translate_polygon(self,
                                 polygon: Polygon,
                                 cosine: Scalar,
                                 sine: Scalar,
                                 step_x: Scalar,
                                 step_y: Scalar,
                                 contour_cls: Type[Contour],
                                 point_cls: Type[Point],
                                 polygon_cls: Type[Polygon]) -> Polygon:
        return polygon_cls(
                self.rotate_translate_contour(polygon.border, cosine, sine,
                                              step_x, step_y, contour_cls,
                                              point_cls),
                [self.rotate_translate_contour(hole, cosine, sine, step_x,
                                               step_y, contour_cls, point_cls)
                 for hole in polygon.holes])

    def rotate_translate_segment(self,
                                 segment: Segment,
                                 cosine: Scalar,
                                 sine: Scalar,
                                 step_x: Scalar,
                                 step_y: Scalar,
                                 point_cls: Type[Point],
                                 segment_cls: Type[Segment]) -> Segment:
        return segment_cls(self.rotate_translate_point(segment.start, cosine,
                                                       sine, step_x, step_y,
                                                       point_cls),
                           self.rotate_translate_point(segment.end, cosine,
                                                       sine, step_x, step_y,
                                                       point_cls))

    __slots__ = ('_point_to_step', '_rotate_point_around_origin',
                 '_rotate_translate_point')

    def __init__(self,
                 point_to_step: PointStep,
                 rotate_point_around_origin: PointRotatorAroundOrigin,
                 rotate_translate_point: PointTranslatingRotator) -> None:
        self._point_to_step = point_to_step
        self._rotate_point_around_origin = rotate_point_around_origin
        self._rotate_translate_point = rotate_translate_point

    __repr__ = generate_repr(__init__,
                             with_module_name=True)


exact_context = Context(
        point_to_step=exact.point_to_step,
        rotate_point_around_origin=exact.rotate_point_around_origin,
        rotate_translate_point=exact.rotate_translate_point)
plain_context = Context(
        point_to_step=plain.point_to_step,
        rotate_point_around_origin=plain.rotate_point_around_origin,
        rotate_translate_point=plain.rotate_translate_point)
robust_context = Context(
        point_to_step=robust.point_to_step,
        rotate_point_around_origin=robust.rotate_point_around_origin,
        rotate_translate_point=robust.rotate_translate_point)



from cfractions import Fraction
from shewchuk import Expansion

from ground.core.hints import (Contour,
                               Scalar)
from ground.core.robust import to_cross_product


def signed_area(contour: Contour[Scalar],
                *,
                _half: Fraction = Fraction(1, 2)) -> Scalar:
    vertices = contour.vertices
    result, vertex = Expansion(), vertices[-1]
    for next_vertex in vertices:
        result = result + to_cross_product(vertex.x, vertex.y, next_vertex.x,
                                           next_vertex.y)
        vertex = next_vertex
    return result * _half



from ground.core.hints import (Point,
                               Scalar)
from ground.core.primitive import rationalize


def point_squared_distance(first: Point, second: Point) -> Scalar:
    return ((rationalize(first.x) - rationalize(second.x)) ** 2
            + (rationalize(first.y) - rationalize(second.y)) ** 2)



from shewchuk import vectors_dot_product

from ground.core.hints import (Point,
                               Scalar)


def multiply(first_start: Point,
             first_end: Point,
             second_start: Point,
             second_end: Point) -> Scalar:
    return vectors_dot_product(first_start.x, first_start.y, first_end.x,
                               first_end.y, second_start.x, second_start.y,
                               second_end.x, second_end.y)



from typing import (Callable,
                    Type)

from shewchuk import Expansion

from ground.core.hints import (Multisegment,
                               Point,
                               Scalar)
from ground.core.robust import to_squared_points_distance


def centroid(multisegment: Multisegment,
             point_cls: Type[Point],
             sqrt: Callable[[Scalar], Scalar]) -> Point:
    accumulated_x = accumulated_y = accumulated_length = Expansion()
    for segment in multisegment.segments:
        start, end = segment.start, segment.end
        length = sqrt(to_squared_points_distance(end.x, end.y, start.x,
                                                 start.y))
        accumulated_x += (start.x + end.x) * length
        accumulated_y += (start.y + end.y) * length
        accumulated_length += length
    inverted_divisor = 1 / (2 * accumulated_length)
    return point_cls(accumulated_x * inverted_divisor,
                     accumulated_y * inverted_divisor)



from typing import (Callable,
                    Type)

from reprit.base import generate_repr

from ground.core.enums import (Orientation,
                               Relation)
from ground.core.hints import (Point,
                               TernaryPointFunction)
from . import (exact,
               plain)

CollisionDetector = Callable[[Point, Point, Point, Point,
                              TernaryPointFunction], bool]
ContainmentChecker = Callable[[Point, Point, Point,
                               TernaryPointFunction[Orientation]], bool]
Intersector = Callable[[Point, Point, Point, Point, Type[Point],
                        TernaryPointFunction[bool]], Point]
Relater = Callable[[Point, Point, Point, Point,
                    TernaryPointFunction[Orientation]], Relation]


class Context:
    @staticmethod
    def collision_detector(first_start: Point,
                           first_end: Point,
                           second_start: Point,
                           second_end: Point,
                           orienteer: TernaryPointFunction[Orientation]
                           ) -> bool:
        second_start_orientation = orienteer(first_start, first_end,
                                             second_start)
        if (second_start_orientation is Orientation.COLLINEAR
                and _bounding_box_contains_point(first_start, first_end,
                                                 second_end)):
            return True
        second_end_orientation = orienteer(first_start, first_end, second_end)
        if (second_end_orientation is Orientation.COLLINEAR
                and _bounding_box_contains_point(first_start, first_end,
                                                 second_end)):
            return True
        first_start_orientation = orienteer(second_start, second_end,
                                            first_start)
        if (first_start_orientation is Orientation.COLLINEAR
                and _bounding_box_contains_point(second_start, second_end,
                                                 first_end)):
            return True
        first_end_orientation = orienteer(second_start, second_end, first_end)
        return (first_end_orientation is Orientation.COLLINEAR
                and _bounding_box_contains_point(second_start, second_end,
                                                 first_end)
                or (second_start_orientation is not second_end_orientation
                    and first_start_orientation is not first_end_orientation))

    @staticmethod
    def containment_checker(start: Point,
                            end: Point,
                            point: Point,
                            orienteer: TernaryPointFunction[Orientation]
                            ) -> bool:
        return (point == start or point == end
                or (_bounding_box_contains_point(start, end, point)
                    and orienteer(start, end, point) is Orientation.COLLINEAR))

    @staticmethod
    def relater(test_start: Point,
                test_end: Point,
                goal_start: Point,
                goal_end: Point,
                orienteer: TernaryPointFunction[Orientation]) -> Relation:
        if test_start > test_end:
            test_start, test_end = test_end, test_start
        if goal_start > goal_end:
            goal_start, goal_end = goal_end, goal_start
        starts_equal = test_start == goal_start
        ends_equal = test_end == goal_end
        if starts_equal and ends_equal:
            return Relation.EQUAL
        test_start_orientation = orienteer(goal_end, goal_start, test_start)
        test_end_orientation = orienteer(goal_end, goal_start, test_end)
        if (test_start_orientation
                is not Orientation.COLLINEAR
                is not test_end_orientation):
            if (test_start_orientation > 0) is (test_end_orientation > 0):
                return Relation.DISJOINT
            else:
                goal_start_orientation = orienteer(test_start, test_end,
                                                   goal_start)
                goal_end_orientation = orienteer(test_start, test_end,
                                                 goal_end)
                if (goal_start_orientation
                        is not Orientation.COLLINEAR
                        is not goal_end_orientation):
                    return (Relation.DISJOINT
                            if ((goal_start_orientation > 0)
                                is (goal_end_orientation > 0))
                            else Relation.CROSS)
                elif goal_start_orientation is not Orientation.COLLINEAR:
                    return (Relation.TOUCH
                            if test_start < goal_end < test_end
                            else Relation.DISJOINT)
                elif goal_end_orientation is not Orientation.COLLINEAR:
                    return (Relation.TOUCH
                            if test_start < goal_start < test_end
                            else Relation.DISJOINT)
        elif test_start_orientation is not Orientation.COLLINEAR:
            return (Relation.TOUCH
                    if goal_start <= test_end <= goal_end
                    else Relation.DISJOINT)
        elif test_end_orientation is not Orientation.COLLINEAR:
            return (Relation.TOUCH
                    if goal_start <= test_start <= goal_end
                    else Relation.DISJOINT)
        elif starts_equal:
            return (Relation.COMPONENT
                    if test_end < goal_end
                    else Relation.COMPOSITE)
        elif ends_equal:
            return (Relation.COMPOSITE
                    if test_start < goal_start
                    else Relation.COMPONENT)
        elif test_start == goal_end or test_end == goal_start:
            return Relation.TOUCH
        elif goal_start < test_start < goal_end:
            return (Relation.COMPONENT
                    if test_end < goal_end
                    else Relation.OVERLAP)
        elif test_start < goal_start < test_end:
            return (Relation.COMPOSITE
                    if goal_end < test_end
                    else Relation.OVERLAP)
        else:
            return Relation.DISJOINT

    @property
    def intersector(self) -> Intersector:
        return self._intersector

    __slots__ = '_intersector',

    def __init__(self, intersector: Intersector) -> None:
        self._intersector = intersector

    __repr__ = generate_repr(__init__,
                             with_module_name=True)


exact_context = Context(exact.intersect)
plain_context = Context(plain.intersect)


def _bounding_box_contains_point(start: Point, end: Point, point: Point
                                 ) -> bool:
    start_x, start_y, end_x, end_y = start.x, start.y, end.x, end.y
    min_x, max_x = (start_x, end_x) if start_x < end_x else (end_x, start_x)
    min_y, max_y = (start_y, end_y) if start_y < end_y else (end_y, start_y)
    return min_x <= point.x <= max_x and min_y <= point.y <= max_y



from typing import (Iterable,
                    Tuple,
                    Type)

from .hints import (Box,
                    Contour,
                    Point,
                    Polygon,
                    Scalar,
                    Segment)


def from_contour(contour: Contour, box_cls: Type[Box]) -> Box:
    return from_points(contour.vertices, box_cls)


def from_contours(contours: Iterable[Contour], box_cls: Type[Box]) -> Box:
    iterator = iter(contours)
    contour = next(iterator)
    min_x, max_x, min_y, max_y = _from_points(contour.vertices)
    for contour in iterator:
        (contour_min_x, contour_max_x, contour_min_y,
         contour_max_y) = _from_points(contour.vertices)
        if max_x < contour_max_x:
            max_x = contour_max_x
        if contour_min_x < min_x:
            min_x = contour_min_x
        if max_y < contour_max_y:
            max_y = contour_max_y
        if contour_min_y < min_y:
            min_y = contour_min_y
    return box_cls(min_x, max_x, min_y, max_y)


def from_points(points: Iterable[Point], box_cls: Type[Box]) -> Box:
    return box_cls(*_from_points(points))


def from_polygon(polygon: Polygon, box_cls: Type[Box]) -> Box:
    return from_points(polygon.border.vertices, box_cls)


def from_polygons(polygons: Iterable[Polygon], box_cls: Type[Box]) -> Box:
    return from_contours((polygon.border for polygon in polygons), box_cls)


def from_segment(segment: Segment, box_cls: Type[Box]) -> Box:
    return box_cls(*_from_segment(segment))


def from_segments(segments: Iterable[Segment], box_cls: Type[Box]) -> Box:
    iterator = iter(segments)
    segment = next(iterator)
    min_x, max_x, min_y, max_y = _from_segment(segment)
    for segment in iterator:
        (segment_min_x, segment_max_x, segment_min_y,
         segment_max_y) = _from_segment(segment)
        if max_x < segment_max_x:
            max_x = segment_max_x
        if segment_min_x < min_x:
            min_x = segment_min_x
        if max_y < segment_max_y:
            max_y = segment_max_y
        if segment_min_y < min_y:
            min_y = segment_min_y
    return box_cls(min_x, max_x, min_y, max_y)


def _from_points(points: Iterable[Point]
                 ) -> Tuple[Scalar, Scalar, Scalar, Scalar]:
    iterator = iter(points)
    point = next(iterator)
    max_x = min_x = point.x
    max_y = min_y = point.y
    for point in iterator:
        if max_x < point.x:
            max_x = point.x
        elif point.x < min_x:
            min_x = point.x
        if max_y < point.y:
            max_y = point.y
        elif point.y < min_y:
            min_y = point.y
    return min_x, max_x, min_y, max_y


def _from_segment(segment: Segment) -> Tuple[Scalar, Scalar, Scalar, Scalar]:
    start, end = segment.start, segment.end
    max_x, min_x = (end.x, start.x) if start.x < end.x else (start.x, end.x)
    max_y, min_y = (end.y, start.y) if start.y < end.y else (start.y, end.y)
    return min_x, max_x, min_y, max_y



from typing import (Sequence,
                    Tuple,
                    Type)

from cfractions import Fraction
from shewchuk import Expansion

from ground.core.hints import (Contour,
                               Point)
from ground.core.robust import to_cross_product


def centroid(contour: Contour,
             point_cls: Type[Point],
             third: Fraction = Fraction(1, 3)) -> Point:
    x_numerator, y_numerator, double_area = centroid_components(
            contour.vertices)
    inverted_denominator = third / double_area
    return point_cls(x_numerator * inverted_denominator,
                     y_numerator * inverted_denominator)


def centroid_components(vertices: Sequence[Point]
                        ) -> Tuple[Expansion, Expansion, Expansion]:
    double_area = x_numerator = y_numerator = Expansion()
    prev = vertices[-1]
    prev_x, prev_y = prev.x, prev.y
    for vertex in vertices:
        x, y = vertex.x, vertex.y
        area_component = to_cross_product(prev_x, prev_y, x, y)
        double_area = double_area + area_component
        x_numerator = x_numerator + area_component * (prev_x + x)
        y_numerator = y_numerator + area_component * (prev_y + y)
        prev_x, prev_y = x, y
    return x_numerator, y_numerator, double_area



from typing import Type

from cfractions import Fraction

from ground.core.hints import (Point,
                               Segment)


def centroid(segment: Segment, point_cls: Type[Point],
             *,
             _half: Fraction = Fraction(1, 2)) -> Point:
    return point_cls(_half * (segment.start.x + segment.end.x),
                     _half * (segment.start.y + segment.end.y))



from ground.core.enums import (Kind,
                               Orientation)
from ground.core.hints import (Point,
                               QuaternaryPointFunction,
                               Scalar)
from ground.core.primitive import to_sign
from ground.core.vector.exact import (cross,
                                      dot)


def kind(vertex: Point,
         first_ray_point: Point,
         second_ray_point: Point,
         dot_producer: QuaternaryPointFunction[Scalar] = dot.multiply) -> Kind:
    return Kind(to_sign(dot_producer(vertex, first_ray_point, vertex,
                                     second_ray_point)))


def orientation(vertex: Point,
                first_ray_point: Point,
                second_ray_point: Point,
                cross_producer: QuaternaryPointFunction[Scalar]
                = cross.multiply) -> Orientation:
    return Orientation(to_sign(cross_producer(vertex, first_ray_point, vertex,
                                              second_ray_point)))



from cfractions import Fraction

from .hints import Scalar


def robust_divide(dividend: Scalar, divisor: Scalar) -> Scalar:
    return (dividend / Fraction(divisor)
            if isinstance(divisor, int)
            else dividend / divisor)



from typing import (Callable,
                    Type)

from reprit.base import generate_repr

from ground.core.hints import (Contour,
                               Multipoint,
                               Multipolygon,
                               Multisegment,
                               Point,
                               Polygon,
                               Scalar,
                               Segment)
from . import (exact,
               plain,
               robust)

PointTranslator = Callable[[Point, Scalar, Scalar, Type[Point]], Point]


class Context:
    @property
    def translate_point(self) -> PointTranslator:
        return self._translate_point

    def translate_contour(self,
                          contour: Contour,
                          step_x: Scalar,
                          step_y: Scalar,
                          contour_cls: Type[Contour],
                          point_cls: Type[Point]) -> Contour:
        return contour_cls([self.translate_point(vertex, step_x, step_y,
                                                 point_cls)
                            for vertex in contour.vertices])

    def translate_multipoint(self,
                             multipoint: Multipoint,
                             step_x: Scalar,
                             step_y: Scalar,
                             multipoint_cls: Type[Multipoint],
                             point_cls: Type[Point]) -> Multipoint:
        return multipoint_cls([self.translate_point(point, step_x, step_y,
                                                    point_cls)
                               for point in multipoint.points])

    def translate_multipolygon(self,
                               multipolygon: Multipolygon,
                               step_x: Scalar,
                               step_y: Scalar,
                               contour_cls: Type[Contour],
                               multipolygon_cls: Type[Multipolygon],
                               point_cls: Type[Point],
                               polygon_cls: Type[Polygon]) -> Multipolygon:
        return multipolygon_cls(
                [self.translate_polygon(polygon, step_x, step_y, contour_cls,
                                        point_cls, polygon_cls)
                 for polygon in multipolygon.polygons])

    def translate_multisegment(self,
                               multisegment: Multisegment,
                               step_x: Scalar,
                               step_y: Scalar,
                               multisegment_cls: Type[Multisegment],
                               point_cls: Type[Point],
                               segment_cls: Type[Segment]) -> Multisegment:
        return multisegment_cls([self.translate_segment(segment, step_x,
                                                        step_y, point_cls,
                                                        segment_cls)
                                 for segment in multisegment.segments])

    def translate_polygon(self,
                          polygon: Polygon,
                          step_x: Scalar,
                          step_y: Scalar,
                          contour_cls: Type[Contour],
                          point_cls: Type[Point],
                          polygon_cls: Type[Polygon]) -> Polygon:
        return polygon_cls(self.translate_contour(polygon.border, step_x,
                                                  step_y, contour_cls,
                                                  point_cls),
                           [self.translate_contour(hole, step_x, step_y,
                                                   contour_cls, point_cls)
                            for hole in polygon.holes])

    def translate_segment(self,
                          segment: Segment,
                          step_x: Scalar,
                          step_y: Scalar,
                          point_cls: Type[Point],
                          segment_cls: Type[Segment]) -> Segment:
        return segment_cls(self.translate_point(segment.start, step_x, step_y,
                                                point_cls),
                           self.translate_point(segment.end, step_x, step_y,
                                                point_cls))

    __slots__ = '_translate_point',

    def __init__(self, translate_point: PointTranslator) -> None:
        self._translate_point = translate_point

    __repr__ = generate_repr(__init__,
                             with_module_name=True)


exact_context = Context(exact.translate_point)
plain_context = Context(plain.translate_point)
robust_context = Context(robust.translate_point)



from itertools import groupby
from typing import (Iterable,
                    List,
                    Sequence,
                    TypeVar)

from .enums import Orientation
from .hints import (Point,
                    TernaryPointFunction)


def to_convex_hull(points: Sequence[Point],
                   orienteer: TernaryPointFunction[Orientation]
                   ) -> List[Point]:
    points = _to_unique_just_seen(sorted(points))
    lower, upper = (_to_sub_hull(points, orienteer),
                    _to_sub_hull(reversed(points), orienteer))
    return lower[:-1] + upper[:-1] or points


_T = TypeVar('_T')


def _to_unique_just_seen(iterable: Iterable[_T]) -> Sequence[_T]:
    return [key for key, _ in groupby(iterable)]


def _to_sub_hull(points: Iterable[Point],
                 orienteer: TernaryPointFunction[Orientation]) -> List[Point]:
    result = []
    for point in points:
        while len(result) >= 2:
            if (orienteer(result[-2], result[-1], point)
                    is not Orientation.COUNTERCLOCKWISE):
                del result[-1]
            else:
                break
        result.append(point)
    return result



from typing import (Callable,
                    Type)

from ground.core.hints import (Contour,
                               Point,
                               Scalar)
from ground.core.primitive import rationalize


def centroid(contour: Contour,
             point_cls: Type[Point],
             sqrt: Callable[[Scalar], Scalar]) -> Point:
    vertices = contour.vertices
    accumulated_x = accumulated_y = accumulated_length = 0
    vertex = vertices[-1]
    start_x, start_y = rationalize(vertex.x), rationalize(vertex.y)
    for vertex in vertices:
        end_x, end_y = rationalize(vertex.x), rationalize(vertex.y)
        length = sqrt((end_x - start_x) ** 2 + (end_y - start_y) ** 2)
        accumulated_x += (start_x + end_x) * length
        accumulated_y += (start_y + end_y) * length
        accumulated_length += length
        start_x, start_y = end_x, end_y
    inverted_divisor = 1 / (2 * accumulated_length)
    return point_cls(accumulated_x * inverted_divisor,
                     accumulated_y * inverted_divisor)



from cfractions import Fraction

from ground.core.hints import (Contour,
                               Scalar)


def signed_area(contour: Contour[Scalar],
                *,
                _half: Fraction = Fraction(1, 2)) -> Scalar:
    vertices = contour.vertices
    result, vertex = 0, vertices[-1]
    for next_vertex in vertices:
        result += vertex.x * next_vertex.y - next_vertex.x * vertex.y
        vertex = next_vertex
    return _half * result



from typing import (Callable,
                    Type)

from reprit.base import generate_repr

from ground.core.hints import (Contour,
                               Multipoint,
                               Multipolygon,
                               Multisegment,
                               Point,
                               Polygon,
                               Segment,
                               SquareRooter)
from .exact import (contour as exact_contour,
                    multipoint as exact_multipoint,
                    multipolygon as exact_multipolygon,
                    multisegment as exact_multisegment,
                    polygon as exact_polygon,
                    region as exact_region,
                    segment as exact_segment)
from .plain import (contour as plain_contour,
                    multipoint as plain_multipoint,
                    multipolygon as plain_multipolygon,
                    multisegment as plain_multisegment,
                    polygon as plain_polygon,
                    region as plain_region,
                    segment as plain_segment)
from .robust import (contour as robust_contour,
                     multipoint as robust_multipoint,
                     multipolygon as robust_multipolygon,
                     multisegment as robust_multisegment,
                     polygon as robust_polygon,
                     region as robust_region,
                     segment as robust_segment)

ContourCentroid = Callable[[Contour, Type[Point], SquareRooter], Point]
MultipointCentroid = Callable[[Multipoint, Type[Point]], Point]
MultipolygonCentroid = Callable[[Multipolygon, Type[Point]], Point]
MultisegmentCentroid = Callable[[Multisegment, Type[Point], SquareRooter],
                                Point]
PolygonCentroid = Callable[[Polygon, Type[Point]], Point]
RegionCentroid = Callable[[Contour, Type[Point]], Point]
SegmentCentroid = Callable[[Segment, Type[Point]], Point]


class Context:
    __slots__ = ('_contour_centroid', '_multipoint_centroid',
                 '_multipolygon_centroid', '_multisegment_centroid',
                 '_polygon_centroid', '_region_centroid', '_segment_centroid')

    def __init__(self,
                 contour_centroid: ContourCentroid,
                 multipoint_centroid: MultipointCentroid,
                 multipolygon_centroid: MultipolygonCentroid,
                 multisegment_centroid: MultisegmentCentroid,
                 polygon_centroid: PolygonCentroid,
                 region_centroid: RegionCentroid,
                 segment_centroid: SegmentCentroid) -> None:
        self._contour_centroid = contour_centroid
        self._multipoint_centroid = multipoint_centroid
        self._multipolygon_centroid = multipolygon_centroid
        self._multisegment_centroid = multisegment_centroid
        self._polygon_centroid = polygon_centroid
        self._region_centroid = region_centroid
        self._segment_centroid = segment_centroid

    __repr__ = generate_repr(__init__,
                             with_module_name=True)

    @property
    def contour_centroid(self) -> ContourCentroid:
        return self._contour_centroid

    @property
    def multipoint_centroid(self) -> MultipointCentroid:
        return self._multipoint_centroid

    @property
    def multipolygon_centroid(self) -> MultipolygonCentroid:
        return self._multipolygon_centroid

    @property
    def multisegment_centroid(self) -> MultisegmentCentroid:
        return self._multisegment_centroid

    @property
    def polygon_centroid(self) -> PolygonCentroid:
        return self._polygon_centroid

    @property
    def region_centroid(self) -> RegionCentroid:
        return self._region_centroid

    @property
    def segment_centroid(self) -> SegmentCentroid:
        return self._segment_centroid


exact_context = Context(contour_centroid=exact_contour.centroid,
                        multipoint_centroid=exact_multipoint.centroid,
                        multipolygon_centroid=exact_multipolygon.centroid,
                        multisegment_centroid=exact_multisegment.centroid,
                        polygon_centroid=exact_polygon.centroid,
                        region_centroid=exact_region.centroid,
                        segment_centroid=exact_segment.centroid)
plain_context = Context(contour_centroid=plain_contour.centroid,
                        multipoint_centroid=plain_multipoint.centroid,
                        multipolygon_centroid=plain_multipolygon.centroid,
                        multisegment_centroid=plain_multisegment.centroid,
                        polygon_centroid=plain_polygon.centroid,
                        region_centroid=plain_region.centroid,
                        segment_centroid=plain_segment.centroid)
robust_context = Context(contour_centroid=robust_contour.centroid,
                         multipoint_centroid=robust_multipoint.centroid,
                         multipolygon_centroid=robust_multipolygon.centroid,
                         multisegment_centroid=robust_multisegment.centroid,
                         polygon_centroid=robust_polygon.centroid,
                         region_centroid=robust_region.centroid,
                         segment_centroid=robust_segment.centroid)



from typing import Callable

from cfractions import Fraction

from ground.core.enums import Relation
from ground.core.hints import (Point,
                               QuaternaryPointFunction,
                               Scalar)
from .point import point_squared_distance as point_point_squared_distance


def point_squared_distance(start: Point,
                           end: Point,
                           point: Point,
                           dot_producer: QuaternaryPointFunction[Scalar],
                           inverse: Callable[[Scalar], Scalar]
                           = Fraction(1).__truediv__) -> Scalar:
    end_factor = max(0, min(1,
                            dot_producer(start, point, start, end)
                            * inverse(point_point_squared_distance(start,
                                                                   end))))
    start_factor = 1 - end_factor
    return ((start_factor * start.x + end_factor * end.x - point.x) ** 2
            + (start_factor * start.y + end_factor * end.y - point.y) ** 2)


def segment_squared_distance(first_start: Point,
                             first_end: Point,
                             second_start: Point,
                             second_end: Point,
                             dot_producer: QuaternaryPointFunction[Scalar],
                             segments_relater
                             : QuaternaryPointFunction[Relation]) -> Scalar:
    return (min(point_squared_distance(first_start, first_end, second_start,
                                       dot_producer),
                point_squared_distance(first_start, first_end, second_end,
                                       dot_producer),
                point_squared_distance(second_start, second_end, first_start,
                                       dot_producer),
                point_squared_distance(second_start, second_end, first_end,
                                       dot_producer))
            if segments_relater(first_start, first_end, second_start,
                                second_end) is Relation.DISJOINT
            else 0)



from typing import Callable

from reprit.base import generate_repr

from ground.core.hints import (Contour,
                               Scalar)
from .exact import region as exact_region
from .plain import region as plain_region
from .robust import region as robust_region

RegionSignedMeasure = Callable[[Contour[Scalar]], Scalar]


class Context:
    @property
    def region_signed_area(self) -> RegionSignedMeasure:
        return self._region_signed_area

    __slots__ = '_region_signed_area',

    def __init__(self,
                 region_signed_area: RegionSignedMeasure) -> None:
        self._region_signed_area = region_signed_area

    __repr__ = generate_repr(__init__,
                             with_module_name=True)


exact_context = Context(exact_region.signed_area)
plain_context = Context(plain_region.signed_area)
robust_context = Context(robust_region.signed_area)



from typing import (Callable,
                    Type)

from reprit.base import generate_repr

from ground.core.hints import (Box,
                               Point,
                               QuaternaryPointFunction,
                               Scalar,
                               Segment)
from .exact import (box as exact_box,
                    point as exact_point,
                    segment as exact_segment)
from .plain import (box as plain_box,
                    point as plain_point,
                    segment as plain_segment)
from .robust import (box as robust_box,
                     point as robust_point,
                     segment as robust_segment)

BoxPointMetric = Callable[[Box, Point], Scalar]
BoxSegmentMetric = Callable[[Box, Segment, QuaternaryPointFunction[Scalar],
                             QuaternaryPointFunction[bool], Type[Point]],
                            Scalar]
PointPointMetric = Callable[[Point, Point], Scalar]
SegmentPointMetric = Callable[[Point, Point, Point,
                               QuaternaryPointFunction[Scalar]],
                              Scalar]
SegmentSegmentMetric = Callable[[Point, Point, Point, Point,
                                 QuaternaryPointFunction[Scalar],
                                 QuaternaryPointFunction[bool]], Scalar]


class Context:
    __slots__ = ('_box_point_squared_metric', '_box_segment_squared_metric',
                 '_point_point_squared_metric',
                 '_segment_point_squared_metric',
                 '_segment_segment_squared_metric')

    def __init__(self,
                 *,
                 box_point_squared_metric: BoxPointMetric,
                 box_segment_squared_metric: BoxSegmentMetric,
                 point_point_squared_metric: PointPointMetric,
                 segment_point_squared_metric: SegmentPointMetric,
                 segment_segment_squared_metric: SegmentSegmentMetric
                 ) -> None:
        self._box_point_squared_metric = box_point_squared_metric
        self._box_segment_squared_metric = box_segment_squared_metric
        self._point_point_squared_metric = point_point_squared_metric
        self._segment_point_squared_metric = segment_point_squared_metric
        self._segment_segment_squared_metric = segment_segment_squared_metric

    __repr__ = generate_repr(__init__,
                             with_module_name=True)

    @property
    def box_point_squared_metric(self) -> BoxPointMetric:
        return self._box_point_squared_metric

    @property
    def box_segment_squared_metric(self) -> BoxSegmentMetric:
        return self._box_segment_squared_metric

    @property
    def point_point_squared_metric(self) -> PointPointMetric:
        return self._point_point_squared_metric

    @property
    def segment_point_squared_metric(self) -> SegmentPointMetric:
        return self._segment_point_squared_metric

    @property
    def segment_segment_squared_metric(self) -> SegmentSegmentMetric:
        return self._segment_segment_squared_metric


exact_context = Context(
        box_point_squared_metric=exact_box.point_squared_distance,
        box_segment_squared_metric=exact_box.segment_squared_distance,
        point_point_squared_metric=exact_point.point_squared_distance,
        segment_point_squared_metric=exact_segment.point_squared_distance,
        segment_segment_squared_metric=exact_segment.segment_squared_distance)
plain_context = Context(
        box_point_squared_metric=plain_box.point_squared_distance,
        box_segment_squared_metric=plain_box.segment_squared_distance,
        point_point_squared_metric=plain_point.point_squared_distance,
        segment_point_squared_metric=plain_segment.point_squared_distance,
        segment_segment_squared_metric=plain_segment.segment_squared_distance)
robust_context = Context(
        box_point_squared_metric=robust_box.point_squared_distance,
        box_segment_squared_metric=robust_box.segment_squared_distance,
        point_point_squared_metric=robust_point.point_squared_distance,
        segment_point_squared_metric=robust_segment.point_squared_distance,
        segment_segment_squared_metric=robust_segment.segment_squared_distance)



from typing import (Callable,
                    Type)

from cfractions import Fraction

from ground.core.hints import (Multipoint,
                               Point)


def centroid(multipoint: Multipoint,
             point_cls: Type[Point],
             inverse: Callable[[int], Fraction] = Fraction(1).__truediv__
             ) -> Point:
    result_x = result_y = 0
    for point in multipoint.points:
        result_x += point.x
        result_y += point.y
    inverted_points_count = inverse(len(multipoint.points))
    return point_cls(result_x * inverted_points_count,
                     result_y * inverted_points_count)



from cfractions import Fraction

from ground.core.hints import (Contour,
                               Scalar)
from ground.core.primitive import rationalize


def signed_area(contour: Contour[Scalar],
                *,
                _half: Fraction = Fraction(1, 2)) -> Scalar:
    vertices = contour.vertices
    result, vertex = 0, vertices[-1]
    vertex_x, vertex_y = rationalize(vertex.x), rationalize(vertex.y)
    for next_vertex in vertices:
        next_vertex_x, next_vertex_y = (rationalize(next_vertex.x),
                                        rationalize(next_vertex.y))
        result += vertex_x * next_vertex_y - next_vertex_x * vertex_y
        vertex_x, vertex_y = next_vertex_x, next_vertex_y
    return _half * result



from typing import Type

from ground.core.hints import (Point,
                               Scalar)


def translate_point(point: Point,
                    step_x: Scalar,
                    step_y: Scalar,
                    point_cls: Type[Point]) -> Point:
    return point_cls(point.x + step_x, point.y + step_y)



from typing import (Tuple,
                    Type)

from ground.core.hints import (Point,
                               Scalar)
from ground.core.primitive import rationalize


def rotate_point_around_origin(point: Point,
                               cosine: Scalar,
                               sine: Scalar,
                               point_cls: Type[Point]) -> Point:
    cosine, sine = rationalize(cosine), rationalize(sine)
    x, y = rationalize(point.x), rationalize(point.y)
    return point_cls(cosine * x - sine * y, sine * x + cosine * y)


def rotate_translate_point(point: Point,
                           cosine: Scalar,
                           sine: Scalar,
                           step_x: Scalar,
                           step_y: Scalar,
                           point_cls: Type[Point]) -> Point:
    x, y = rationalize(point.x), rationalize(point.y)
    cosine, sine = rationalize(cosine), rationalize(sine)
    return point_cls(cosine * x - sine * y + rationalize(step_x),
                     sine * x + cosine * y + rationalize(step_y))


def point_to_step(point: Point,
                  cosine: Scalar,
                  sine: Scalar) -> Tuple[Scalar, Scalar]:
    x, y = rationalize(point.x), rationalize(point.y)
    cosine, sine = rationalize(cosine), rationalize(sine)
    return x - (cosine * x - sine * y), y - (sine * x + cosine * y)



from abc import abstractmethod
from numbers import Real
from typing import (Callable,
                    Sequence,
                    TypeVar,
                    Union)

from symba.base import Expression

try:
    from typing import (Protocol,
                        runtime_checkable)
except ImportError:
    from typing_extensions import (Protocol,
                                   runtime_checkable)

Scalar = TypeVar('Scalar', Expression, Real)
SquareRooter = Callable[[Scalar], Scalar]


@runtime_checkable
class Point(Protocol[Scalar]):
    """
    **Point** is a minimal element of the plane
    defined by pair of real numbers (called *point's coordinates*).

    Points considered to be sorted lexicographically,
    with abscissas being compared first.
    """
    __slots__ = ()

    @abstractmethod
    def __new__(cls, x: Scalar, y: Scalar) -> 'Point':
        """Constructs point given its coordinates."""

    @property
    @abstractmethod
    def x(self) -> Scalar:
        """Abscissa of the point."""

    @property
    @abstractmethod
    def y(self) -> Scalar:
        """Ordinate of the point."""

    @abstractmethod
    def __eq__(self, other: 'Point') -> bool:
        """Checks if the point is equal to the other."""

    @abstractmethod
    def __ge__(self, other: 'Point') -> bool:
        """Checks if the point is greater than or equal to the other."""

    @abstractmethod
    def __gt__(self, other: 'Point') -> bool:
        """Checks if the point is greater than the other."""

    @abstractmethod
    def __hash__(self) -> int:
        """Returns hash value of the point."""

    @abstractmethod
    def __le__(self, other: 'Point') -> bool:
        """Checks if the point is less than or equal to the other."""

    @abstractmethod
    def __lt__(self, other: 'Point') -> bool:
        """Checks if the point is less than the other."""


Range = TypeVar('Range')
QuaternaryPointFunction = Callable[[Point, Point, Point, Point], Range]
TernaryPointFunction = Callable[[Point, Point, Point], Range]


@runtime_checkable
class Box(Protocol[Scalar]):
    """
    **Box** is a limited closed region
    defined by axis-aligned rectangular contour.
    """
    __slots__ = ()

    @abstractmethod
    def __new__(cls,
                min_x: Scalar,
                max_x: Scalar,
                min_y: Scalar,
                max_y: Scalar) -> 'Box':
        """Constructs box given its coordinates limits."""

    @property
    @abstractmethod
    def max_x(self) -> Scalar:
        """Maximum ``x``-coordinate of the box."""

    @property
    @abstractmethod
    def max_y(self) -> Scalar:
        """Maximum ``y``-coordinate of the box."""

    @property
    @abstractmethod
    def min_x(self) -> Scalar:
        """Minimum ``x``-coordinate of the box."""

    @property
    @abstractmethod
    def min_y(self) -> Scalar:
        """Minimum ``y``-coordinate of the box."""

    @abstractmethod
    def __eq__(self, other: 'Box') -> bool:
        """Checks if the box is equal to the other."""


@runtime_checkable
class Empty(Protocol):
    """Represents an empty set of points."""
    __slots__ = ()

    @abstractmethod
    def __new__(cls) -> 'Empty':
        """Constructs empty geometry."""

    @abstractmethod
    def __eq__(self, other: 'Empty') -> bool:
        """Checks if the empty geometry is equal to the other."""


_T = TypeVar('_T')
Maybe = Union[Empty, _T]


@runtime_checkable
class Multipoint(Protocol[Scalar]):
    """
    **Multipoint** is a discrete geometry
    that represents non-empty set of unique points.
    """
    __slots__ = ()

    @abstractmethod
    def __new__(cls, points: Sequence[Point]) -> 'Multipoint':
        """Constructs multipoint given its points."""

    @property
    @abstractmethod
    def points(self) -> Sequence[Point]:
        """Points of the multipoint."""

    @abstractmethod
    def __eq__(self, other: 'Multipoint') -> bool:
        """Checks if the multipoint is equal to the other."""


@runtime_checkable
class Segment(Protocol[Scalar]):
    """
    **Segment** (or **line segment**) is a linear geometry that represents
    a limited continuous part of the line containing more than one point
    defined by a pair of unequal points (called *segment's endpoints*).
    """
    __slots__ = ()

    @abstractmethod
    def __new__(cls, start: Point, end: Point) -> 'Segment':
        """Constructs segment given its endpoints."""

    @property
    @abstractmethod
    def start(self) -> Point:
        """Start endpoint of the segment."""

    @property
    @abstractmethod
    def end(self) -> Point:
        """End endpoint of the segment."""

    @abstractmethod
    def __eq__(self, other: 'Segment') -> bool:
        """Checks if the segment is equal to the other."""


@runtime_checkable
class Multisegment(Protocol[Scalar]):
    """
    **Multisegment** is a linear geometry that represents set of two or more
    non-crossing and non-overlapping segments.
    """
    __slots__ = ()

    @abstractmethod
    def __new__(cls, segments: Sequence[Segment]) -> 'Multisegment':
        """Constructs multisegment given its segments."""

    @property
    @abstractmethod
    def segments(self) -> Sequence[Segment]:
        """Segments of the multisegment."""

    @abstractmethod
    def __eq__(self, other: 'Multisegment') -> bool:
        """Checks if the multisegment is equal to the other."""


@runtime_checkable
class Contour(Protocol[Scalar]):
    """
    **Contour** is a linear geometry that represents closed simple polyline
    defined by a sequence of points (called *contour's vertices*).
    """
    __slots__ = ()

    @abstractmethod
    def __new__(cls, vertices: Sequence[Point]) -> 'Contour':
        """Constructs contour given its vertices."""

    @property
    @abstractmethod
    def vertices(self) -> Sequence[Point]:
        """Vertices of the contour."""

    @abstractmethod
    def __eq__(self, other: 'Contour') -> bool:
        """Checks if the contour is equal to the other."""


@runtime_checkable
class Polygon(Protocol[Scalar]):
    """
    **Polygon** is a shaped geometry that represents limited closed region
    defined by the pair of outer contour (called *polygon's border*)
    and possibly empty sequence of inner contours (called *polygon's holes*).
    """
    __slots__ = ()

    @abstractmethod
    def __new__(cls, border: Contour, holes: Sequence[Contour]) -> 'Polygon':
        """Constructs polygon given its border and holes."""

    @property
    @abstractmethod
    def border(self) -> Contour:
        """Border of the polygon."""

    @property
    @abstractmethod
    def holes(self) -> Sequence[Contour]:
        """Holes of the polygon."""

    @abstractmethod
    def __eq__(self, other: 'Polygon') -> bool:
        """Checks if the polygon is equal to the other."""


@runtime_checkable
class Multipolygon(Protocol[Scalar]):
    """
    **Multipolygon** is a shaped geometry that represents set of two or more
    non-overlapping polygons intersecting only in discrete set of points.
    """
    __slots__ = ()

    @abstractmethod
    def __new__(cls, polygons: Sequence[Polygon]) -> 'Multipolygon':
        """Constructs multipolygon given its polygons."""

    @property
    @abstractmethod
    def polygons(self) -> Sequence[Polygon]:
        """Polygons of the multipolygon."""

    @abstractmethod
    def __eq__(self, other: 'Multipolygon') -> bool:
        """Checks if the multipolygon is equal to the other."""


Linear = Union[Segment, Multisegment, Contour]
Shaped = Union[Polygon, Multipolygon]


@runtime_checkable
class Mix(Protocol[Scalar]):
    """
    **Mix** is a set of two or more non-empty geometries
    with different dimensions.
    """
    __slots__ = ()

    @abstractmethod
    def __new__(cls,
                discrete: Maybe[Multipoint],
                linear: Maybe[Linear],
                shaped: Maybe[Shaped]) -> 'Mix':
        """Constructs mix given its components."""

    @abstractmethod
    def __eq__(self, other: 'Mix') -> bool:
        """Checks if the mix is equal to the other."""

    @property
    @abstractmethod
    def discrete(self) -> Maybe[Multipoint]:
        """Discrete component of the mix."""

    @property
    @abstractmethod
    def linear(self) -> Maybe[Linear]:
        """Linear component of the mix."""

    @property
    @abstractmethod
    def shaped(self) -> Maybe[Shaped]:
        """Shaped component of the mix."""



from typing import Type

from ground.core.hints import (Point,
                               Scalar)
from ground.core.primitive import rationalize


def scale_point(point: Point,
                factor_x: Scalar,
                factor_y: Scalar,
                point_cls: Type[Point]) -> Point:
    return point_cls(rationalize(point.x) * rationalize(factor_x),
                     rationalize(point.y) * rationalize(factor_y))



from reprit.base import generate_repr

from ground.core.enums import (Kind,
                               Orientation)
from ground.core.hints import (QuaternaryPointFunction)
from .exact import (kind as exact_kind,
                    orientation as exact_orientation)
from .plain import (kind as plain_kind,
                    orientation as plain_orientation)
from .robust import (kind as robust_kind,
                     orientation as robust_orientation)


class Context:
    __slots__ = '_kind', '_orientation'

    def __init__(self,
                 kind: QuaternaryPointFunction[Kind],
                 orientation: QuaternaryPointFunction[Orientation]) -> None:
        self._kind, self._orientation = kind, orientation

    __repr__ = generate_repr(__init__,
                             with_module_name=True)

    @property
    def kind(self) -> QuaternaryPointFunction[Kind]:
        return self._kind

    @property
    def orientation(self) -> QuaternaryPointFunction[Orientation]:
        return self._orientation


exact_context = Context(kind=exact_kind,
                        orientation=exact_orientation)
plain_context = Context(kind=plain_kind,
                        orientation=plain_orientation)
robust_context = Context(kind=robust_kind,
                         orientation=robust_orientation)



from shewchuk import (kind as raw_kind,
                      orientation as raw_orientation)

from ground.core.enums import (Kind,
                               Orientation)
from ground.core.hints import Point


def kind(vertex: Point,
         first_ray_point: Point,
         second_ray_point: Point) -> Kind:
    return Kind(raw_kind(vertex.x, vertex.y, first_ray_point.x,
                         first_ray_point.y, second_ray_point.x,
                         second_ray_point.y))


def orientation(vertex: Point,
                first_ray_point: Point,
                second_ray_point: Point) -> Orientation:
    return Orientation(raw_orientation(vertex.x, vertex.y, first_ray_point.x,
                                       first_ray_point.y, second_ray_point.x,
                                       second_ray_point.y))



from shewchuk import incircle_test

from ground.core.enums import Location
from ground.core.hints import Point


def test(point: Point,
         first: Point,
         second: Point,
         third: Point) -> Location:
    return Location(1 + incircle_test(point.x, point.y, first.x, first.y,
                                      second.x, second.y, third.x, third.y))



import sys
from collections import OrderedDict
from itertools import chain
from typing import (Callable,
                    Iterable,
                    Type,
                    Union)

from reprit.base import generate_repr

from ground.core.hints import (Contour,
                               Empty,
                               Mix,
                               Multipoint,
                               Multipolygon,
                               Multisegment,
                               Point,
                               Polygon,
                               Scalar,
                               Segment)
from ground.core.packing import (pack_mix,
                                 pack_points,
                                 pack_segments)
from . import (exact,
               plain,
               robust)

PointScaler = Callable[[Point, Scalar, Scalar, Type[Point]], Point]


class Context:
    @property
    def scale_point(self) -> PointScaler:
        return self._scale_point

    def scale_contour(self,
                      contour: Contour,
                      factor_x: Scalar,
                      factor_y: Scalar,
                      contour_cls: Type[Contour],
                      multipoint_cls: Type[Multipoint],
                      point_cls: Type[Point],
                      segment_cls: Type[Segment]
                      ) -> Union[Contour, Multipoint, Segment]:
        return (self.scale_contour_non_degenerate(contour, factor_x, factor_y,
                                                  contour_cls, point_cls)
                if factor_x and factor_y
                else self.scale_contour_degenerate(contour, factor_x, factor_y,
                                                   multipoint_cls, point_cls,
                                                   segment_cls))

    def scale_contour_non_degenerate(self,
                                     contour: Contour,
                                     factor_x: Scalar,
                                     factor_y: Scalar,
                                     contour_cls: Type[Contour],
                                     point_cls: Type[Point]) -> Contour:
        return contour_cls([self.scale_point(vertex, factor_x, factor_y,
                                             point_cls)
                            for vertex in contour.vertices])

    def scale_contour_degenerate(self,
                                 contour: Contour,
                                 factor_x: Scalar,
                                 factor_y: Scalar,
                                 multipoint_cls: Type[Multipoint],
                                 point_cls: Type[Point],
                                 segment_cls: Type[Segment]
                                 ) -> Union[Multipoint, Segment]:
        return self.scale_vertices_degenerate(contour.vertices, factor_x,
                                              factor_y, multipoint_cls,
                                              point_cls, segment_cls)

    def scale_multipoint(self,
                         multipoint: Multipoint,
                         factor_x: Scalar,
                         factor_y: Scalar,
                         multipoint_cls: Type[Multipoint],
                         point_cls: Type[Point]) -> Multipoint:
        return multipoint_cls(
                [self.scale_point(point, factor_x, factor_y, point_cls)
                 for point in multipoint.points]
                if factor_x and factor_y
                else
                (list(unique_ever_seen(self.scale_point(point, factor_x,
                                                        factor_y, point_cls)
                                       for point in multipoint.points))
                 if factor_x or factor_y
                 else [point_cls(factor_x, factor_y)]))

    def scale_multipolygon(self,
                           multipolygon: Multipolygon,
                           factor_x: Scalar,
                           factor_y: Scalar,
                           contour_cls: Type[Contour],
                           multipoint_cls: Type[Multipoint],
                           multipolygon_cls: Type[Multipolygon],
                           multisegment_cls: Type[Multisegment],
                           point_cls: Type[Point],
                           polygon_cls: Type[Polygon],
                           segment_cls: Type[Segment]
                           ) -> Union[Multipoint, Multipolygon, Multisegment]:
        return (
            multipolygon_cls(
                    [self.scale_polygon_non_degenerate(polygon, factor_x,
                                                       factor_y, contour_cls,
                                                       point_cls, polygon_cls)
                     for polygon in multipolygon.polygons])
            if factor_x and factor_y
            else
            (multisegment_cls([self.scale_polygon_degenerate(
                    polygon, factor_x, factor_y, multipoint_cls, point_cls,
                    segment_cls)
                for polygon in multipolygon.polygons])
             if factor_x or factor_y
             else multipoint_cls([point_cls(factor_x, factor_y)])))

    def scale_multisegment(self,
                           multisegment: Multisegment,
                           factor_x: Scalar,
                           factor_y: Scalar,
                           empty: Empty,
                           mix_cls: Type[Mix],
                           multipoint_cls: Type[Multipoint],
                           multisegment_cls: Type[Multisegment],
                           point_cls: Type[Point],
                           segment_cls: Type[Segment]
                           ) -> Union[Mix, Multipoint, Multisegment, Segment]:
        scaled_points, scaled_segments = [], []
        for segment in multisegment.segments:
            if ((factor_x or not is_segment_horizontal(segment)) and factor_y
                    or factor_x and not is_segment_vertical(segment)):
                scaled_segments.append(self.scale_segment_non_degenerate(
                        segment, factor_x, factor_y, point_cls, segment_cls))
            else:
                scaled_points.append(self.scale_point(segment.start, factor_x,
                                                      factor_y, point_cls))
        return pack_mix(pack_points(list(unique_ever_seen(scaled_points)),
                                    empty, multipoint_cls),
                        pack_segments(scaled_segments, empty,
                                      multisegment_cls),
                        empty, empty, mix_cls)

    def scale_polygon(self,
                      polygon: Polygon,
                      factor_x: Scalar,
                      factor_y: Scalar,
                      contour_cls: Type[Contour],
                      multipoint_cls: Type[Multipoint],
                      point_cls: Type[Point],
                      polygon_cls: Type[Polygon],
                      segment_cls: Type[Segment]
                      ) -> Union[Multipoint, Polygon, Segment]:
        return (self.scale_polygon_non_degenerate(polygon, factor_x, factor_y,
                                                  contour_cls, point_cls,
                                                  polygon_cls)
                if factor_x and factor_y
                else self.scale_polygon_degenerate(polygon, factor_x, factor_y,
                                                   multipoint_cls, point_cls,
                                                   segment_cls))

    def scale_polygon_degenerate(self,
                                 polygon: Polygon,
                                 factor_x: Scalar,
                                 factor_y: Scalar,
                                 multipoint_cls: Type[Multipoint],
                                 point_cls: Type[Point],
                                 segment_cls: Type[Segment]
                                 ) -> Union[Multipoint, Segment]:
        return self.scale_contour_degenerate(polygon.border, factor_x,
                                             factor_y, multipoint_cls,
                                             point_cls, segment_cls)

    def scale_polygon_non_degenerate(self,
                                     polygon: Polygon,
                                     factor_x: Scalar,
                                     factor_y: Scalar,
                                     contour_cls: Type[Contour],
                                     point_cls: Type[Point],
                                     polygon_cls: Type[Polygon]) -> Polygon:
        return polygon_cls(
                self.scale_contour_non_degenerate(polygon.border, factor_x,
                                                  factor_y, contour_cls,
                                                  point_cls),
                [self.scale_contour_non_degenerate(hole, factor_x, factor_y,
                                                   contour_cls, point_cls)
                 for hole in polygon.holes])

    def scale_segment(self,
                      segment: Segment,
                      factor_x: Scalar,
                      factor_y: Scalar,
                      multipoint_cls: Type[Multipoint],
                      point_cls: Type[Point],
                      segment_cls: Type[Segment]
                      ) -> Union[Multipoint, Segment]:
        return (self.scale_segment_non_degenerate(segment, factor_x, factor_y,
                                                  point_cls, segment_cls)
                if ((factor_x or not is_segment_horizontal(segment))
                    and factor_y
                    or factor_x and not is_segment_vertical(segment))
                else multipoint_cls([self.scale_point(segment.start, factor_x,
                                                      factor_y, point_cls)]))

    def scale_segment_non_degenerate(self,
                                     segment: Segment,
                                     factor_x: Scalar,
                                     factor_y: Scalar,
                                     point_cls: Type[Point],
                                     segment_cls: Type[Segment]) -> Segment:
        return segment_cls(self.scale_point(segment.start, factor_x, factor_y,
                                            point_cls),
                           self.scale_point(segment.end, factor_x, factor_y,
                                            point_cls))

    def scale_vertices_degenerate(self,
                                  vertices: Iterable[Point],
                                  factor_x: Scalar,
                                  factor_y: Scalar,
                                  multipoint_cls: Type[Multipoint],
                                  point_cls: Type[Point],
                                  segment_cls: Type[Segment]
                                  ) -> Union[Segment, Multipoint]:
        return (self.scale_vertices_projecting_on_ox(vertices, factor_x,
                                                     factor_y, point_cls,
                                                     segment_cls)
                if factor_x
                else (self.scale_vertices_projecting_on_oy(vertices, factor_x,
                                                           factor_y, point_cls,
                                                           segment_cls)
                      if factor_y
                      else multipoint_cls([point_cls(factor_x, factor_y)])))

    @staticmethod
    def scale_vertices_projecting_on_ox(vertices: Iterable[Point],
                                        factor_x: Scalar,
                                        factor_y: Scalar,
                                        point_cls: Type[Point],
                                        segment_cls: Type[Segment]) -> Segment:
        vertices = iter(vertices)
        min_x = max_x = next(vertices).x
        for vertex in vertices:
            if min_x > vertex.x:
                min_x = vertex.x
            elif max_x < vertex.x:
                max_x = vertex.x
        return segment_cls(point_cls(min_x * factor_x, factor_y),
                           point_cls(max_x * factor_x, factor_y))

    @staticmethod
    def scale_vertices_projecting_on_oy(vertices: Iterable[Point],
                                        factor_x: Scalar,
                                        factor_y: Scalar,
                                        point_cls: Type[Point],
                                        segment_cls: Type[Segment]) -> Segment:
        vertices = iter(vertices)
        min_y = max_y = next(vertices).y
        for vertex in vertices:
            if min_y > vertex.y:
                min_y = vertex.y
            elif max_y < vertex.y:
                max_y = vertex.y
        return segment_cls(point_cls(factor_x, min_y * factor_y),
                           point_cls(factor_x, max_y * factor_y))

    __slots__ = '_scale_point',

    def __init__(self, scale_point: PointScaler) -> None:
        self._scale_point = scale_point

    __repr__ = generate_repr(__init__,
                             with_module_name=True)


exact_context = Context(exact.scale_point)
plain_context = Context(plain.scale_point)
robust_context = Context(robust.scale_point)


def is_segment_horizontal(segment: Segment) -> bool:
    return segment.start.y == segment.end.y


def is_segment_vertical(segment: Segment) -> bool:
    return segment.start.x == segment.end.x


flatten = chain.from_iterable
unique_ever_seen = (dict.fromkeys
                    if sys.version_info >= (3, 6)
                    else OrderedDict.fromkeys)



from typing import (Sequence,
                    Tuple,
                    Type)

from ground.core.hints import (Multipolygon,
                               Point,
                               Polygon,
                               Scalar)
from .polygon import centroid_components as polygon_centroid_components


def centroid(multipolygon: Multipolygon,
             point_cls: Type[Point]) -> Point:
    x_numerator, y_numerator, double_area = centroid_components(
            multipolygon.polygons)
    inverted_denominator = 1 / (3 * double_area)
    return point_cls(x_numerator * inverted_denominator,
                     y_numerator * inverted_denominator)


def centroid_components(polygons: Sequence[Polygon]
                        ) -> Tuple[Scalar, Scalar, Scalar]:
    iterator = iter(polygons)
    polygon = next(iterator)
    x_numerator, y_numerator, double_area = polygon_centroid_components(
            polygon.border, polygon.holes)
    for polygon in iterator:
        (polygon_x_numerator, polygon_y_numerator,
         polygon_double_area) = polygon_centroid_components(polygon.border,
                                                            polygon.holes)
        x_numerator += polygon_x_numerator
        y_numerator += polygon_y_numerator
        double_area += polygon_double_area
    return x_numerator, y_numerator, double_area



from ground.core.hints import (Point,
                               Scalar)
from ground.core.robust import to_squared_points_distance


def point_squared_distance(first: Point, second: Point) -> Scalar:
    return to_squared_points_distance(first.x, first.y, second.x, second.y)



from numbers import (Rational,
                     Real)
from typing import Type

from cfractions import Fraction

from .hints import (Point,
                    Scalar)


def rationalize(value: Scalar) -> Scalar:
    try:
        return Fraction(value)
    except TypeError:
        return value


def square(value: Scalar) -> Scalar:
    return value * value


def to_rational_point(point: Point[Real],
                      point_cls: Type[Point]) -> Point[Rational]:
    return point_cls(rationalize(point.x), rationalize(point.y))


def to_sign(value: Scalar) -> int:
    return 1 if value > 0 else (-1 if value else 0)



from typing import Type

from shewchuk import Expansion

from ground.core.hints import (Point,
                               Segment)


def centroid(segment: Segment, point_cls: Type[Point]) -> Point:
    return point_cls(Expansion(segment.start.x, segment.end.x) / 2,
                     Expansion(segment.start.y, segment.end.y) / 2)



from ground.core.hints import (Point,
                               Scalar)
from ground.core.primitive import rationalize


def multiply(first_start: Point,
             first_end: Point,
             second_start: Point,
             second_end: Point) -> Scalar:
    return ((rationalize(first_end.x) - rationalize(first_start.x))
            * (rationalize(second_end.x) - rationalize(second_start.x))
            + (rationalize(first_end.y) - rationalize(first_start.y))
            * (rationalize(second_end.y) - rationalize(second_start.y)))



from typing import (Sequence,
                    Type,
                    Union)

from .hints import (Empty,
                    Linear,
                    Maybe,
                    Mix,
                    Multipoint,
                    Multisegment,
                    Point,
                    Segment,
                    Shaped)


def pack_mix(discrete: Maybe[Multipoint],
             linear: Maybe[Linear],
             shaped: Maybe[Shaped],
             empty: Empty,
             mix_cls: Type[Mix]
             ) -> Union[Empty, Linear, Mix, Multipoint, Shaped]:
    return (mix_cls(discrete, linear, shaped)
            if (((discrete is not empty) + (linear is not empty)
                 + (shaped is not empty)) >= 2)
            else (discrete
                  if discrete is not empty
                  else (linear
                        if linear is not empty
                        else shaped)))


def pack_points(points: Sequence[Point],
                empty: Empty,
                multipoint_cls: Type[Multipoint]) -> Maybe[Multipoint]:
    return multipoint_cls(points) if points else empty


def pack_segments(segments: Sequence[Segment],
                  empty: Empty,
                  multisegment_cls: Type[Multisegment]
                  ) -> Union[Empty, Multisegment, Segment]:
    return ((multisegment_cls(segments)
             if len(segments) > 1
             else segments[0])
            if segments
            else empty)



from typing import (Sequence,
                    Tuple,
                    Type)

from ground.core.hints import (Contour,
                               Point,
                               Scalar)
from ground.core.primitive import rationalize


def centroid(contour: Contour, point_cls: Type[Point]) -> Point:
    x_numerator, y_numerator, double_area = centroid_components(
            contour.vertices)
    inverted_divisor = 1 / (3 * double_area)
    return point_cls(x_numerator * inverted_divisor,
                     y_numerator * inverted_divisor)


def centroid_components(vertices: Sequence[Point]
                        ) -> Tuple[Scalar, Scalar, Scalar]:
    double_area = x_numerator = y_numerator = 0
    prev_vertex = vertices[-1]
    prev_x, prev_y = rationalize(prev_vertex.x), rationalize(prev_vertex.y)
    for vertex in vertices:
        x, y = rationalize(vertex.x), rationalize(vertex.y)
        area_component = prev_x * y - prev_y * x
        double_area += area_component
        x_numerator += (prev_x + x) * area_component
        y_numerator += (prev_y + y) * area_component
        prev_x, prev_y = x, y
    return x_numerator, y_numerator, double_area



from shewchuk import Expansion

from .hints import Scalar
from .primitive import square


def to_cross_product(first_x: Scalar,
                     first_y: Scalar,
                     second_x: Scalar,
                     second_y: Scalar) -> Expansion:
    """
    Returns expansion of vectors' cross product.
    """
    return Expansion(first_x) * second_y - Expansion(second_x) * first_y


def to_squared_points_distance(first_x: Scalar,
                               first_y: Scalar,
                               second_x: Scalar,
                               second_y: Scalar) -> Expansion:
    return (square(Expansion(first_x, -second_x))
            + square(Expansion(first_y, -second_y)))



from typing import Type

from shewchuk import Expansion

from ground.core.hints import (Point,
                               Scalar)


def scale_point(point: Point,
                factor_x: Scalar,
                factor_y: Scalar,
                point_cls: Type[Point]) -> Point:
    return point_cls(Expansion(point.x) * factor_x,
                     Expansion(point.y) * factor_y)



from typing import Type

from ground.core.hints import (Point,
                               Scalar)
from ground.core.primitive import rationalize


def translate_point(point: Point,
                    step_x: Scalar,
                    step_y: Scalar,
                    point_cls: Type[Point]) -> Point:
    return point_cls(rationalize(point.x) + rationalize(step_x),
                     rationalize(point.y) + rationalize(step_y))



from operator import eq
from typing import (Sequence,
                    TypeVar)

from reprit.base import generate_repr

from . import hints


class Point:
    __slots__ = '_x', '_y'

    def __init__(self, x: hints.Scalar, y: hints.Scalar) -> None:
        self._x, self._y = x, y

    @property
    def x(self) -> hints.Scalar:
        return self._x

    @property
    def y(self) -> hints.Scalar:
        return self._y

    def __eq__(self, other: 'Point') -> bool:
        return (self.x == other.x and self.y == other.y
                if isinstance(other, Point)
                else NotImplemented)

    def __hash__(self) -> int:
        return hash((self.x, self.y))

    def __le__(self, other: 'Point') -> bool:
        return (self.x < other.x or self.x == other.x and self.y <= other.y
                if isinstance(other, Point)
                else NotImplemented)

    def __lt__(self, other: 'Point') -> bool:
        return (self.x < other.x or self.x == other.x and self.y < other.y
                if isinstance(other, Point)
                else NotImplemented)

    __repr__ = generate_repr(__init__)


class Empty:
    __slots__ = ()

    def __init__(self) -> None:
        pass

    def __eq__(self, other: 'Empty'):
        return isinstance(other, Empty) or NotImplemented

    __repr__ = generate_repr(__init__)


class Multipoint:
    __slots__ = '_points',

    def __init__(self, points: Sequence[hints.Point]) -> None:
        self._points = points

    @property
    def points(self) -> Sequence[hints.Point]:
        return self._points

    def __eq__(self, other: 'Multipoint') -> bool:
        return (are_sequences_equivalent(self.points, other.points)
                if isinstance(other, Multipoint)
                else NotImplemented)

    __repr__ = generate_repr(__init__)


class Segment:
    __slots__ = '_start', '_end'

    def __init__(self, start: hints.Point, end: hints.Point) -> None:
        self._start, self._end = start, end

    @property
    def start(self) -> hints.Point:
        return self._start

    @property
    def end(self) -> hints.Point:
        return self._end

    def __eq__(self, other: 'Segment') -> bool:
        return (self.start == other.start and self.end == other.end
                or self.start == other.end and self.end == other.start
                if isinstance(other, Segment)
                else NotImplemented)

    __repr__ = generate_repr(__init__)


class Multisegment:
    __slots__ = '_segments',

    def __init__(self, segments: Sequence[hints.Segment]) -> None:
        self._segments = segments

    @property
    def segments(self) -> Sequence[hints.Segment]:
        return self._segments

    def __eq__(self, other: 'Multisegment') -> bool:
        return (are_sequences_equivalent(self.segments, other.segments)
                if isinstance(other, Multisegment)
                else NotImplemented)

    __repr__ = generate_repr(__init__)


class Contour:
    __slots__ = '_vertices',

    def __init__(self, vertices: Sequence[hints.Polygon]) -> None:
        self._vertices = vertices

    @property
    def vertices(self) -> Sequence[hints.Polygon]:
        return self._vertices

    def __eq__(self, other: 'Contour') -> bool:
        return (are_sequences_equivalent(self.vertices, other.vertices)
                if isinstance(other, Contour)
                else NotImplemented)

    __repr__ = generate_repr(__init__)


class Box:
    __slots__ = '_min_x', '_max_x', '_min_y', '_max_y'

    def __init__(self,
                 min_x: hints.Scalar,
                 max_x: hints.Scalar,
                 min_y: hints.Scalar,
                 max_y: hints.Scalar) -> None:
        self._min_x, self._max_x, self._min_y, self._max_y = (min_x, max_x,
                                                              min_y, max_y)

    @property
    def max_x(self) -> hints.Scalar:
        return self._max_x

    @property
    def max_y(self) -> hints.Scalar:
        return self._max_y

    @property
    def min_x(self) -> hints.Scalar:
        return self._min_x

    @property
    def min_y(self) -> hints.Scalar:
        return self._min_y

    def __eq__(self, other: 'Box') -> bool:
        return (self.min_x == other.min_x and self.max_x == other.max_x
                and self.min_y == other.min_y and self.max_y == other.max_y
                if isinstance(other, Box)
                else NotImplemented)

    __repr__ = generate_repr(__init__)


class Polygon:
    __slots__ = '_border', '_holes'

    def __init__(self, border: hints.Contour, holes: Sequence[hints.Contour]
                 ) -> None:
        self._border, self._holes = border, holes

    @property
    def border(self) -> hints.Contour:
        return self._border

    @property
    def holes(self) -> Sequence[hints.Contour]:
        return self._holes

    def __eq__(self, other: 'Polygon') -> bool:
        return (self.border == other.border
                and are_sequences_equivalent(self.holes, other.holes)
                if isinstance(other, Polygon)
                else NotImplemented)

    __repr__ = generate_repr(__init__)


class Multipolygon:
    __slots__ = '_polygons',

    def __init__(self, polygons: Sequence[hints.Polygon]) -> None:
        self._polygons = polygons

    @property
    def polygons(self) -> Sequence[hints.Polygon]:
        return self._polygons

    def __eq__(self, other: 'Multipolygon') -> bool:
        return (are_sequences_equivalent(self.polygons, other.polygons)
                if isinstance(other, Multipolygon)
                else NotImplemented)

    __repr__ = generate_repr(__init__)


class Mix:
    __slots__ = '_discrete', '_linear', '_shaped'

    def __init__(self,
                 discrete: hints.Maybe[hints.Multipoint],
                 linear: hints.Maybe[hints.Linear],
                 shaped: hints.Maybe[hints.Shaped]) -> None:
        self._discrete, self._linear, self._shaped = discrete, linear, shaped

    @property
    def discrete(self) -> hints.Maybe[hints.Multipoint]:
        return self._discrete

    @property
    def linear(self) -> hints.Maybe[hints.Linear]:
        return self._linear

    @property
    def shaped(self) -> hints.Maybe[hints.Shaped]:
        return self._shaped

    def __eq__(self, other: 'Mix') -> bool:
        return (self.discrete == other.discrete
                and self.linear == other.linear
                and self.shaped == other.shaped
                if isinstance(other, Mix)
                else NotImplemented)

    __repr__ = generate_repr(__init__)


_T = TypeVar('_T')


def are_sequences_equivalent(left: Sequence[_T], right: Sequence[_T]) -> bool:
    return len(left) == len(right) and all(map(eq, left, right))



from ground.core.hints import (Point,
                               QuaternaryPointFunction,
                               Scalar)
from ground.core.primitive import rationalize
from .point import point_squared_distance as point_point_squared_distance


def point_squared_distance(segment_start: Point,
                           segment_end: Point,
                           point: Point,
                           dot_producer: QuaternaryPointFunction[Scalar]
                           ) -> Scalar:
    end_factor = max(0, min(1,
                            dot_producer(segment_start, point, segment_start,
                                         segment_end)
                            / point_point_squared_distance(segment_start,
                                                           segment_end)))
    start_factor = 1 - end_factor
    return ((start_factor * rationalize(segment_start.x)
             + end_factor * rationalize(segment_end.x)
             - rationalize(point.x)) ** 2
            + (start_factor * rationalize(segment_start.y)
               + end_factor * rationalize(segment_end.y)
               - rationalize(point.y)) ** 2)


def segment_squared_distance(first_start: Point,
                             first_end: Point,
                             second_start: Point,
                             second_end: Point,
                             dot_producer: QuaternaryPointFunction[Scalar],
                             segments_collision_detector
                             : QuaternaryPointFunction[bool]) -> Scalar:
    return (0
            if segments_collision_detector(first_start, first_end,
                                           second_start, second_end)
            else min(point_squared_distance(first_start, first_end,
                                            second_start, dot_producer),
                     point_squared_distance(first_start, first_end, second_end,
                                            dot_producer),
                     point_squared_distance(second_start, second_end,
                                            first_start, dot_producer),
                     point_squared_distance(second_start, second_end,
                                            first_end, dot_producer)))



from ground.core.hints import (Point,
                               Scalar)


def multiply(first_start: Point,
             first_end: Point,
             second_start: Point,
             second_end: Point) -> Scalar:
    return ((first_end.x - first_start.x) * (second_end.y - second_start.y)
            - (first_end.y - first_start.y) * (second_end.x - second_start.x))



from ground.core.hints import (Point,
                               Scalar)


def point_squared_distance(first: Point, second: Point) -> Scalar:
    return (first.x - second.x) ** 2 + (first.y - second.y) ** 2
